﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

[System.Serializable]
public struct TerrainLayerDesc
{
	public string name;
	public float height;
	public Sprite[] textures;
	public int drawLayer;
	public bool notWalkable;
}

[System.Serializable]
public struct TerrainObstacleDesc
{
	public string name;
	public float minHeight;
	public float maxHeight;
	public Sprite[] textures;
	public float chance;
	public int extraSpace;
}

[System.Serializable]
public struct SpawnableDesc
{
	public GameObject prefab;
	public float chance;
	public bool staticThing;
}


[System.Flags]
public enum TileFlags
{
	kNone = 0,
	kNotWalkable = (1<<0)
}

public class WorldTile
{
	public const int kSize = 16;
	public const int kKillFrames = 5;
	int originX, originY;
	GameObject m_RootObject;
	TileFlags[] m_TileFlags;
	int m_LastUsedFrame;
	
	public WorldTile(int ox, int oy)
	{
		originX = ox;
		originY = oy;
		m_RootObject = new GameObject(string.Format ("WorldTile_{0}_{1}", ox, oy));
		m_TileFlags = new TileFlags[kSize*kSize];
		m_LastUsedFrame = Time.frameCount;
	}
	
	public int xMin { get { return originX; } }
	public int yMin { get { return originY; } }
	public int xMax { get { return originX + kSize; } }
	public int yMax { get { return originY + kSize; } }
	public bool Contains(int x, int y)
	{
		return x >= originX && y >= originY && x < xMax && y < yMax;
	}

	public GameObject rootObject { get { return m_RootObject; } }
	public TileFlags[] tileFlags { get { return m_TileFlags; } }

	public int GetIndex(int x, int y)
	{
		x -= originX;
		y -= originY;
		int idx = y*kSize + x;
		if (idx < 0 || idx >= kSize*kSize)
		{
			Debug.LogError(string.Format ("Tile index {0},{1} out of bounds", x, y));
		}
		return idx;
	}

	public bool IsWalkable(int x, int y)
	{
		return (m_TileFlags[GetIndex(x,y)] & TileFlags.kNotWalkable) == 0;
	}

	public void Touch()
	{
		m_LastUsedFrame = Time.frameCount;
	}
	
	public bool DestroyIfOld()
	{
		if (Time.frameCount > m_LastUsedFrame + kKillFrames)
		{
			Object.Destroy (m_RootObject);
			return true;
		}
		return false;
	}
}


public class World : MonoBehaviour
{
	public TerrainLayerDesc[] terrainLayers;
	public TerrainObstacleDesc[] terrainObstacles;
	public SpawnableDesc[] spawnables;
	public Material terrainLayerMaterial;
	
	public static World instance;

	bool m_SkipObjects = false;
	List<WorldTile> m_Tiles = new List<WorldTile>();
	HashFunction m_Hash = new XXHash(System.DateTime.Now.Second);
	int m_CoordOffset = 1024 + System.DateTime.Now.Millisecond;
	
	public void Awake()
	{
		if (instance != null)
		{
			Debug.LogError ("World instance already exists!");
		}
		instance = this;
		Random.seed = System.DateTime.Now.Second + System.DateTime.Now.Minute;
	}

	public void Start()
	{
		Vector3 playerPos = PlayerController.instance.transform.position;
		// don't generate enemies or pickups in initial area around the world
		m_SkipObjects = true;
		EnsureHasTiles ((int)playerPos.x-6, (int)playerPos.y-6, (int)playerPos.x+6, (int)playerPos.y+6);
		m_SkipObjects = false;
	}

	public bool IsWalkable(int x, int y)
	{
		var tile = FindTile (x, y, false);
		if (tile == null)
			return false;
		return tile.IsWalkable (x, y);
	}

	public WorldTile FindTile(int x, int y, bool errorIfNotFound = true)
	{
		for (var i = 0; i < m_Tiles.Count; ++i)
		{
			var tile = m_Tiles[i];
			if (tile.Contains(x, y) && tile.rootObject != null)
			{
				tile.Touch();
				return tile;
			}
		}
		if (errorIfNotFound)
			Debug.LogWarning(string.Format ("No tile found at {0},{1}", x, y), this);
		return null;
	}

	public static int RoundUp(int x, int mul)
	{
		if (x < 0)
			return (x/mul)*mul;
		else
			return ((x+mul-1)/mul)*mul;
	}
	public static int RoundDown(int x, int mul)
	{
		if (x < 0)
			return ((x-mul+1)/mul)*mul;
		else
			return (x/mul)*mul;
	}

	public void EnsureHasTiles(int xmin, int ymin, int xmax, int ymax)
	{
		xmin = RoundDown (xmin, WorldTile.kSize);
		ymin = RoundDown (ymin, WorldTile.kSize);
		xmax = RoundUp   (xmax, WorldTile.kSize);
		ymax = RoundUp   (ymax, WorldTile.kSize);

		// create newly needed tiles
		for (var y = ymin; y <= ymax; y += WorldTile.kSize)
		{
			for (var x = xmin; x <= xmax; x += WorldTile.kSize)
			{
				WorldTile tile = FindTile (x, y, false);
				if (tile == null)
				{
					// generate new tile
					tile = GenerateTile (x, y);
					m_Tiles.Add (tile);
				}
			}
		}
	}

	void LateUpdate()
	{
		CleanupOldTiles();
	}

	void CleanupOldTiles()
	{
		List<WorldTile> newTiles = new List<WorldTile>(m_Tiles.Count);
		for (var i = 0; i < m_Tiles.Count; ++i)
		{
			var tile = m_Tiles[i];
			if (!tile.DestroyIfOld())
				newTiles.Add (tile);
		}
		m_Tiles = newTiles;
	}

	WorldTile GenerateTile(int xo, int yo)
	{
		int x1 = xo;
		int y1 = yo;
		int x2 = x1 + WorldTile.kSize;
		int y2 = y1 + WorldTile.kSize;

		var tile = new WorldTile(x1, y1);

		for (var y = y1; y < y2; ++y)
		{
			for (var x = x1; x < x2; ++x)
			{
				// several octaves of simplex noise
				float height = SimplexNoise.Generate (x*0.01f + m_CoordOffset, y*0.01f + m_CoordOffset);
				height += SimplexNoise.Generate (x*0.047f + m_CoordOffset, y*0.047f + m_CoordOffset) * 0.3f;
				height += SimplexNoise.Generate (x*0.167f + m_CoordOffset, y*0.167f + m_CoordOffset) * 0.15f;

				height = height * 0.5f + 0.5f;
				height = Mathf.Clamp01(height);

				// find suitable terrain layer and create piece for it
				for (var i = terrainLayers.Length-1; i >= 0; --i)
				{
					var layer = terrainLayers[i];
					if (layer.height <= height)
					{
						InstantiateTerrainPiece(tile, i, x, y);
						break;
					}
				}

				// create obstacles
				var hashValue = m_Hash.Value (x, y);
				for (var i = 0; i < terrainObstacles.Length; ++i)
				{
					var chance = hashValue - i * 0.01f;
					var obs = terrainObstacles[i];
					if (height >= obs.minHeight && height <= obs.maxHeight)
					{
						if (chance >= 0.0f && chance < obs.chance)
						{
							if (x-obs.extraSpace >= x1 && y-obs.extraSpace >= y1 && x+obs.extraSpace < x2 && y+obs.extraSpace < y2)
							{
								InstantiateObstaclePiece(tile, i, x, y);
								break;
							}
						}
					}
				}

				// spawn various items & enemies
				if (tile.IsWalkable(x, y) && !m_SkipObjects)
				{
					for (var i = 0; i < spawnables.Length; ++i)
					{
						var chance = hashValue - 0.1f - i * 0.01f;
						if (chance >= 0.0f && chance < spawnables[i].chance)
						{
							GameObject go = Object.Instantiate(spawnables[i].prefab, new Vector3(x, y, 0), Quaternion.identity) as GameObject;
							if (spawnables[i].staticThing)
								go.transform.parent = tile.rootObject.transform;
							break;
						}
					}
				}
			}
		}
		return tile;
	}

	void InstantiateObstaclePiece(WorldTile tile, int obsIndex, int x, int y)
	{
		var go = new GameObject("OstaclePiece" + obsIndex);
		
		// position it
		var tr = go.transform;
		tr.parent = tile.rootObject.transform;
		tr.position = new Vector3(x, y, 0);

		// assign sprite
		var layer = terrainObstacles[obsIndex];
		var sr = go.AddComponent<SpriteRenderer>();
		sr.sprite = layer.textures[Random.Range (0,layer.textures.Length)];
		sr.sortingOrder = 1000;
		
		// set flags
		var space = layer.extraSpace;
		for (var yy = y-space; yy <= y+space; ++yy)
		{
			for (var xx = x-space; xx <= x+space; ++xx)
			{
				var flagsIndex = tile.GetIndex(xx, yy);
				tile.tileFlags[flagsIndex] |= TileFlags.kNotWalkable;
			}
		}
	}

	void InstantiateTerrainPiece(WorldTile tile, int layerIndex, int x, int y)
	{
		var go = new GameObject("TilePiece" + layerIndex);

		// position it
		var tr = go.transform;
		tr.parent = tile.rootObject.transform;
		tr.position = new Vector3(x, y, 0);
		int rotIndex = Random.Range (0,4);
		tr.Rotate (0,0,rotIndex * 90.0f + Random.Range (-2.0f,2.0f));

		// assign sprite
		var layer = terrainLayers[layerIndex];
		var sr = go.AddComponent<SpriteRenderer>();
		sr.sprite = layer.textures[Random.Range (0,layer.textures.Length)];
		sr.material = terrainLayerMaterial;
		//Color col = sr.color;
		//col.a = SimplexNoise.Generate (x*0.1f,y*0.1f) * 0.5f + 0.5f;
		//sr.color = col;
		sr.sortingOrder = layer.drawLayer * 100 + layerIndex;

		// set flags
		var flagsIndex = tile.GetIndex(x, y);
		if (layer.notWalkable)
		{
			tile.tileFlags[flagsIndex] |= TileFlags.kNotWalkable;
		}

		/*
		// create box collider if needed
		if (tile.obstacle)
		{
			var box = go.AddComponent<BoxCollider2D>();
			box.size = new Vector2(1,1);
		}
		*/
	}
}
